import pathlib

import pycape
from pycape.experimental import cli

CAPE_HOST = "https://app.capeprivacy.com"
ECHO_DEPLOY_PATH = pathlib.Path(__file__).parent.absolute() / "echo"


cape = pycape.Cape(url=CAPE_HOST)
# Deploy Cape function
function_ref = cli.deploy(ECHO_DEPLOY_PATH, token_expiry=100)
print("Echo deployed:")
print(f"\t- ID: {function_ref.id}")
print(f"\t- Token: {function_ref.token}")
print(f"\t- Checksum: {function_ref.checksum}")
# Encrypt input
message = cape.encrypt("Welcome to Cape".encode(), token=function_ref.token)
# Run Cape function
result = cape.run(function_ref, message)
print(f"The result is: {result.decode()}")



import io
import logging
import math
import zipfile
from typing import Dict
from typing import List

import cbor2
import requests
from cose.keys import EC2Key
from cose.keys.curves import P384
from cose.messages import Sign1Message
from cryptography.x509 import load_der_x509_certificate
from OpenSSL import crypto

_AWS_ROOT_CERT_ARCHIVE = (
    "https://aws-nitro-enclaves.amazonaws.com/AWS_NitroEnclaves_Root-G1.zip"
)

logger = logging.getLogger("pycape")


def download_root_cert():
    logger.debug(
        f"* Downloading AWS root cert for attestation from {_AWS_ROOT_CERT_ARCHIVE}..."
    )
    r = requests.get(_AWS_ROOT_CERT_ARCHIVE)
    f = zipfile.ZipFile(io.BytesIO(r.content))
    with f.open("root.pem") as p:
        root_cert = p.read()
    logger.debug("AWS root cert received.")
    return root_cert


def parse_attestation(attestation, root_cert):
    logger.debug("* Parsing attestation document...")

    payload = cbor2.loads(attestation)
    doc = cbor2.loads(payload[2])
    _check_wellformed_attestation(
        doc,
        expected_keys=["certificate", "cabundle", "public_key"],
    )
    doc_cert = doc["certificate"]
    cabundle = doc["cabundle"]
    logger.debug("* Attestation document parsed.")

    verify_attestation_signature(attestation, doc_cert)
    verify_cert_chain(root_cert, cabundle, doc_cert)
    return doc


def verify_cert_chain(root_cert, cabundle, cert):
    logger.debug("* Verifying attestation certificate chain...")
    cert = crypto.load_certificate(crypto.FILETYPE_ASN1, cert)

    # Create an X509Store object for the CA bundles
    store = crypto.X509Store()

    # Create the CA cert object from PEM string, and store into X509Store
    _cert = crypto.load_certificate(crypto.FILETYPE_PEM, root_cert)
    store.add_cert(_cert)

    # Get the CA bundle from attestation document and store into X509Store
    # Except the first certificate, which is the root certificate
    chain = []
    for _cert_binary in cabundle:
        _cert = crypto.load_certificate(crypto.FILETYPE_ASN1, _cert_binary)
        chain.append(_cert)

    # Get the X509Store context
    store_ctx = crypto.X509StoreContext(store, cert, chain=chain)

    # Validate the certificate
    # If the cert is invalid, it will raise exception
    store_ctx.verify_certificate()
    logger.debug("* Attestation certificate chain verified.")


def verify_attestation_signature(payload, cert):
    logger.debug("* Verifying attestation certificate signature...")
    cert = load_der_x509_certificate(cert)
    cert_public_numbers = cert.public_key().public_numbers()
    x = cert_public_numbers.x
    y = cert_public_numbers.y

    x = _long_to_bytes(x)
    y = _long_to_bytes(y)

    # Create the EC2 key from public key parameters
    key = EC2Key(x=x, y=y, crv=P384)

    cose_obj = cbor2.loads(payload)
    msg = Sign1Message.from_cose_obj(cose_obj, allow_unknown_attributes=True)
    msg.key = key

    # Verify the signature using the EC2 key
    verified = msg.verify_signature()
    if not verified:
        errmsg = "Malformed attestation doc: incorrect certificate signature."
        logger.error(errmsg)
        raise RuntimeError(errmsg)
    logger.debug("* Attestation certificate signature verified.")


def _long_to_bytes(x: int):
    bytesize = math.ceil(x.bit_length() / 8.0)
    return x.to_bytes(bytesize, byteorder="big")


def _check_wellformed_attestation(doc, expected_keys):
    for doc_key in expected_keys:
        if doc_key not in doc:
            errmsg = (
                "Malformed attestation doc: missing attestation "
                f"document field {doc_key}."
            )
            logger.error(errmsg)
            raise RuntimeError(errmsg)


def verify_pcrs(pcrs: Dict[str, List[str]], doc):
    for key, val in pcrs.items():
        h = doc["pcrs"][int(key)].hex()

        found = False
        for pcr in val:
            if pcr == h:
                found = True
                break

        if not found:
            raise Exception(f"PCR {key} {h} does not match {val}")



import pathlib

import pycape

function_json = pathlib.Path(__file__).parent.absolute() / "echo_token.json"
function_ref = pycape.FunctionRef.from_json(function_json)

cape = pycape.Cape()

# Two options to encrypt for a Cape user

# 1 - retrieve the user's key with cape.key, and pass the key to cape.encrypt
capedocs_key = cape.key(username="capedocs")
encrypted_data = cape.encrypt(b"encrypt against capedocs's key", key=capedocs_key)
result = cape.run(function_ref, encrypted_data)
print(result.decode())

# 2 - pass username directly to cape.encrypt
encrypted_data = cape.encrypt(
    b"quickly encrypt my data for capedocs", username="capedocs"
)
result = cape.run(function_ref, encrypted_data)
print(result.decode())



import json
import os
import pathlib
import subprocess
from typing import Optional
from typing import Union

from pycape import _config as cape_config
from pycape import function_ref as fref
from pycape.cape import _synchronizer


@_synchronizer
async def deploy(
    deploy_path: Union[str, os.PathLike],
    url: Optional[str] = None,
    token_expiry: Optional[str] = None,
    public: bool = False,
) -> fref.FunctionRef:
    """Deploy a directory or a zip file containing a Cape function declared in
    an app.py script.

    This method calls `cape deploy` and `cape token` from the Cape CLI to deploy
    a Cape function then returns a `~.function_ref.FunctionRef` representing
    the deployed function. This `~.function_ref.FunctionRef` will hold a function ID,
    a function token and a function checksum. Note that the ``deploy_path`` has to
    point to a directory or a zip file containing a Cape function declared in an app.py
    file and the size of its content  is currently limited to 1GB.

    Args:
        deploy_path: A path pointing to a directory or a zip file containing
            a Cape function declared in an app.py script.
        url: The Cape platform's websocket URL, which is responsible for forwarding
            client requests to the proper enclave instances. If None, tries to load
            value from the ``CAPE_ENCLAVE_HOST`` environment variable. If no such
            variable value is supplied, defaults to ``"https://app.capeprivacy.com"``.
        token_expriry: Amount of time in seconds until the function token expires.

    Returns:
        A :class:`~.function_ref.FunctionRef` representing the deployed Cape
        function.

    Raises:
        RuntimeError: if the websocket response or the enclave attestation doc is
            malformed, or if the function path is not pointing to a directory
            or a zip file or if folder size exceeds 1GB, or if the Cape CLI cannot
            be found on the device.
    """
    url = url or cape_config.ENCLAVE_HOST
    deploy_path = pathlib.Path(deploy_path)

    cmd_deploy = f"cape deploy {deploy_path} -u {url} -o json"
    if public:
        cmd_deploy += " --public"
    out_deploy, err_deploy = _call_cape_cli(cmd_deploy)
    err_deploy = err_deploy.decode()
    out_deploy = out_deploy.decode()

    err_deploy = err_deploy.split("\n")
    error_output = None

    # Parse err_token to get potential errors
    for msg in err_deploy:
        if "Error" in msg:
            error_output = msg
            error_msg = error_output.partition("Error:")[2]
            raise RuntimeError(f"Cape deploy error - {error_msg}")

    # Parse out_deploy to get function ID and function checksum
    out_deploy = json.loads(out_deploy.split("\n")[0])
    function_id = out_deploy.get("function_id")
    function_checksum = out_deploy.get("function_checksum")

    if function_id is None:
        raise RuntimeError(
            f"Function ID not found in 'cape.deploy' response: \n{err_deploy}"
        )

    function_reference = await token(
        function_id, expiry=token_expiry, function_checksum=function_checksum
    )

    return function_reference


async def token(
    function_id,
    expiry=Optional[None],
    function_checksum: Optional[str] = None,
) -> fref.FunctionRef:
    """Generate a function token (JSON Web Token) based on a function ID.

    This method calls `cape token` from the Cape CLI to generate a function token
    based on a function ID. Tokens can be created statically (long expiration and
    bundled with your application) or created dynamically (short-lived) and have
    an owner specified expiration. This function token is required in addition
    to the function ID when calling a Cape function.

    Args:
        function_id: A function ID string representing a deployed Cape function.
        expiry: Amount of time in seconds until the function token expires.

    Returns:
        A :class:`~.function_ref.FunctionRef` representing the deployed Cape
        function.

        Raises:
        RuntimeError: if the Cape CLI cannot be found on the device.
    """
    if expiry:
        cmd_token = (
            f"cape token {function_id} --function-checksum {function_checksum} "
            f"--expiry {expiry} -o json"
        )
    else:
        cmd_token = (
            f"cape token {function_id} --function-checksum {function_checksum} -o json"
        )

    out_token, err_token = _call_cape_cli(cmd_token)
    err_token = err_token.decode()
    out_token = out_token.decode()

    err_deploy = err_token.split("\n")
    error_output = None
    # Parse err_token to get potential errors
    for i in err_deploy:
        if "Error" in i:
            error_output = i
            error_msg = error_output.partition("Error:")[2]
            raise RuntimeError(f"Cape token error - {error_msg}")

    # Parse out_token to get function token
    function_token = json.loads(out_token.split("\n")[0]).get("function_token")

    if function_token is None:
        raise RuntimeError(
            f"Function token not found in 'cape.token' response: \n{err_deploy}"
        )

    return fref.FunctionRef(function_id, function_token, function_checksum)


def _check_if_cape_cli_available():
    exitcode, output = subprocess.getstatusoutput("cape")
    if exitcode != 0:
        raise RuntimeError(
            f"Please make sure Cape CLI is installed on your device: {output}"
        )


def _call_cape_cli(cape_cmd):
    _check_if_cape_cli_available()
    proc = subprocess.Popen(
        cape_cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err



"""A structured set of metadata representing a deployed Cape function.

A :class:`FunctionRef` is intended to capture any/all metadata related to a Cape
function. The metadata is generally user-supplied, provided to them with the output of
the Cape CLI's ``deploy`` command.

**Usage**

::

    fid = "asdf231lkg1324afdg"
    fchecksum = str(b"2l1h21jhgb2k1jh3".hex())
    fref = FunctionRef(fid, fchecksum)

    cape = Cape()
    cape.connect(fref)
"""
from __future__ import annotations

import json
import os
import pathlib
from typing import Optional
from typing import Union


class FunctionRef:
    """A reference to a Cape function.

    Args:
        id: Required string denoting the function ID of the deployed Cape function.
            Typically given with the output of the Cape CLI's ``deploy`` command.
        token: Required string containing a Cape function token generated by the Cape
            CLI during ``cape token``.
        checksum: Optional string denoting the checksum of the deployed Cape function.
            If supplied as part of a ``FunctionRef``, the :class:`~pycape.cape.Cape`
            client will verify that enclave responses includes a matching checksum
            whenever the ``FunctionRef`` is included in Cape requests.
    """

    def __init__(
        self,
        id: str,
        token: str,
        checksum: Optional[str] = None,
    ):
        id_ = id
        if not isinstance(id_, str):
            raise TypeError(f"Function id must be a string, found {type(id_)}.")
        if not isinstance(token, str):
            raise TypeError(f"Function token must be a string, found {type(token)}.")
        if checksum is not None and not isinstance(checksum, str):
            raise TypeError(
                f"Function checksum must be a string, found {type(checksum)}."
            )
        self._id = id_
        self._checksum = checksum
        self._token = token

    @property
    def id(self):
        return self._id

    @property
    def checksum(self):
        return self._checksum

    @property
    def token(self):
        return self._token

    @classmethod
    def from_json(cls, function_json: Union[str, os.PathLike]) -> FunctionRef:
        """Construct a :class:`~.function_ref.FunctionRef` from a JSON string or file.

        Args:
            function_json: a JSON string or filepath containing function ID, token, and
                optional checksum, e.g. as generated by the Cape CLI ``token`` command.

        Returns:
            A :class:`~.function_ref.FunctionRef` representing the deployed Cape
            function.

        Raises:
            ValueError: if the json token file doesn't exist or, the token file
                doesn't contain a `function_id` or a `function_token`.
        """
        if isinstance(function_json, pathlib.Path):
            function_config = _try_load_json_file(function_json)
            if function_config is None:
                raise ValueError(f"JSON file not found @ {str(function_json)}")

        elif isinstance(function_json, str):
            # try to treat function_json as filepath str
            json_path = pathlib.Path(function_json)
            function_config = _try_load_json_file(json_path)
            # if file not found, treat function_json as json str
            function_config = function_config or json.loads(function_json)

        else:
            raise TypeError(
                "The function_json argument expects a json string or "
                f"a path to a json file, found: {type(function_json)}."
            )

        function_id = function_config.get("function_id")
        if function_id is None:
            raise ValueError("Couldn't find a `function_id` in the token file provided")

        function_token = function_config.get("function_token")
        if function_token is None:
            raise ValueError(
                "Couldn't find a `function_token` in the token file provided"
            )

        function_checksum = function_config.get("function_checksum")

        return cls(function_id, function_token, function_checksum)

    def to_json(self, path: Optional[Union[str, os.PathLike]] = None) -> Optional[str]:
        """Write this :class:`~.function_ref.FunctionRef` to a JSON string or file.

        Args:
            path: Optional file path to write the resulting JSON to.

        Returns:
            If ``path`` is None, a string with this :class:`~.function_ref.FunctionRef`
            as a JSON struct.
        """

        fn_ref_dict = {
            "function_id": self._id,
            "function_token": self._token,
            "function_checksum": self._checksum,
        }

        if path is None:
            return json.dumps(fn_ref_dict)

        with open(path, "w") as f:
            json.dump(fn_ref_dict, f)


def _try_load_json_file(json_file: pathlib.Path):
    if json_file.exists():
        with open(json_file, "r") as f:
            json_output = json.load(f)
        return json_output



import json
import tempfile
import unittest

from pycape import function_ref as fref


class TestFunctionRef(unittest.TestCase):
    def test_from_to_json_string(self):
        exp_function_json = json.dumps(
            {"function_id": "foo", "function_token": "bar", "function_checksum": "yo"}
        )
        function_ref = fref.FunctionRef.from_json(exp_function_json)
        act_function_json = function_ref.to_json()
        assert exp_function_json == act_function_json

    def test_from_to_json_file(self):
        exp_function_ref = fref.FunctionRef("foo", "bar", "yo")
        with tempfile.NamedTemporaryFile() as f:
            exp_function_ref.to_json(f.name)
            act_function_ref = fref.FunctionRef.from_json(f.name)

        assert exp_function_ref.id == act_function_ref.id
        assert exp_function_ref.token == act_function_ref.token
        assert exp_function_ref.checksum == act_function_ref.checksum



import logging

import hybrid_pke

logger = logging.getLogger("pycape")


def encrypt(public_key: bytes, input_bytes: bytes) -> bytes:
    logger.debug("* Encrypting inputs with Hybrid Public Key Encryption (HPKE)")
    hpke = hybrid_pke.default()
    info = b""
    aad = b""
    encap, ciphertext = hpke.seal(public_key, info, aad, input_bytes)
    return encap + ciphertext



import os
import pathlib

from pycape import Cape
from pycape import FunctionRef

if __name__ == "__main__":
    url = os.environ.get("CAPE_HOST", "https://app.capeprivacy.com")
    function_json = os.environ.get("FUNCTION_JSON", "mean_token.json")
    function_json = pathlib.Path(__file__).parent.absolute() / function_json

    function_ref = FunctionRef.from_json(function_json)
    cape = Cape(url=url)

    x = [1, 2, 3, 4]
    result = cape.run(function_ref, x, use_serdio=True)

    print(f"The mean of x is: {result}")



"""The Cape Python client.

The :class:`Cape` client uses websockets to connect to Cape enclaves that are hosting a
user's deployed functions. Before being able to run functions from the Cape client,
users must have gone through the process of developing a Cape function in Python and
deploying it from the CLI and generating a function token.

The majority of the :class:`Cape` client interface can be used in either synchronous or
asynchronous contexts via asyncio.

**Usage**

::

    cape = Cape()

    f = FunctionRef.from_json("function.json")
    cape.connect(f)

    c1 = cape.invoke(3, 4, use_serdio=True)
    print(c1)  # 5

    c2 = cape.invoke(5, 12, use_serdio=True)
    print(c2)  # 13

    cape.close()  # release the enclave connection

    # similar invocation, but async
    c3 = asyncio.run(
        cape.run("9712r5dynf57l1rcns2", 8, 15, use_serdio=True)
    )
    print(c3)  # 17

"""
import base64
import json
import logging
import os
import pathlib
import random
import ssl
import urllib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import requests
import synchronicity
import websockets

import serdio
from pycape import _attestation as attest
from pycape import _config as cape_config
from pycape import _enclave_encrypt as enclave_encrypt
from pycape import cape_encrypt
from pycape import function_ref as fref

logging.basicConfig(format="%(message)s")
_logger = logging.getLogger("pycape")
_synchronizer = synchronicity.Synchronizer(multiwrap_warning=True)


class Cape:
    """A websocket client for interacting with enclaves hosting Cape functions.

    This is the main interface for interacting with Cape functions from Python.
    See module-level documentation :mod:`pycape.cape` for usage example.

    Args:
        url: The Cape platform's websocket URL, which is responsible for forwarding
            client requests to the proper enclave instances. If None, tries to load
            value from the ``CAPE_ENCLAVE_HOST`` environment variable. If no such
            variable value is supplied, defaults to ``"https://app.capeprivacy.com"``.
        verbose: Boolean controlling verbose logging for the ``"pycape"`` logger.
            If True, sets log-level to ``DEBUG``.
    """

    def __init__(
        self,
        url: Optional[str] = None,
        verbose: bool = False,
    ):
        self._url = url or cape_config.ENCLAVE_HOST
        self._root_cert = None
        self._ctx = None

        if verbose:
            _logger.setLevel(logging.DEBUG)

    @_synchronizer
    async def close(self):
        """Closes the current enclave connection."""
        if self._ctx is not None:
            await self._ctx.close()
            self._ctx = None

    @_synchronizer
    async def connect(
        self,
        function_ref: fref.FunctionRef,
        pcrs: Optional[Dict[str, List[str]]] = None,
    ):
        """Connects to the enclave hosting the function denoted by ``function_ref``.

        Note that this method creates a stateful websocket connection, which is a
        necessary precondition for callers of :meth:`~Cape.invoke`. When using the
        default Cape host, the enclave will terminate this websocket connection after
        60s of inactivity. Care should be taken to close the websocket connection with
        :meth:`~Cape.close` once all invocations have finished.

        Args:
            function_ref: A function ID string or :class:`~.function_ref.FunctionRef`
                representing a deployed Cape function.
            pcrs: A dictionary of PCR indexes to a list of potential values.

        Raises:
            RuntimeError: if the websocket response or the enclave attestation doc is
                malformed, or if the enclave fails to return a function checksum
                matching our own.
            Exception: if the enclave threw an error while trying to fulfill the
                connection request.
        """
        function_ref = _check_function_ref(function_ref)
        await self._request_connection(function_ref, pcrs)

    @_synchronizer
    async def encrypt(
        self,
        input: bytes,
        *,
        username: Optional[str] = None,
        token: Optional[str] = None,
        key: Optional[bytes] = None,
        key_path: Optional[Union[str, os.PathLike]] = None,
    ) -> bytes:
        """Encrypts inputs to Cape functions in Cape's encryption format.

        The encrypted value can be used as input to Cape handlers by other callers of
        :meth:`~Cape.invoke` or :meth:`~Cape.run` without giving them plaintext access
        to it. The core encryption functionality uses envelope encryption; the value is
        AES-encrypted with an ephemeral AES key, which is itself encrypted with the Cape
        user's assigned RSA public key. The corresponding RSA private key is only
        accessible from within a Cape enclave, which guarantees secrecy of the encrypted
        value. See the Cape encrypt docs for further detail.

        Args:
            input: Input bytes to encrypt.
            username: A Github username corresponding to a Cape user who's public key
                you want to use for the encryption. See :meth:`~Cape.key` for details.
            token: A Cape token scoped for retrieving a user's Cape public key.
                See :meth:`~Cape.key` for details.
            key: Optional bytes for the Cape key. If None, will delegate to calling
                :meth:`Cape.key` w/ the given ``key_path`` to retrieve the user's Cape
                key.
            key_path: Optional path to a locally-cached Cape key. Used to call
                :meth:`Cape.key` when an explicit ``key`` argument is not provided.

        Returns:
            Tagged ciphertext representing a base64-encoded Cape encryption of the
            ``input``.

        Raises:
            ValueError: if Cape key is not a properly-formatted RSA public key.
            RuntimeError: if the enclave attestation doc does not contain a Cape key,
                if the websocket response or the attestation doc is malformed.
            Exception: if the enclave threw an error while trying to fulfill the
                connection request.
        """
        cape_key = key or await self.key(
            username=username, token=token, key_path=key_path
        )
        ctxt = cape_encrypt.encrypt(input, cape_key)
        # cape-encrypted ctxt must be b64-encoded and tagged
        ctxt = base64.b64encode(ctxt)
        return b"cape:" + ctxt

    @_synchronizer
    @_synchronizer.asynccontextmanager
    async def function_context(
        self,
        function_ref: fref.FunctionRef,
        pcrs: Optional[Dict[str, List[str]]] = None,
    ):
        """Creates a context manager for a given ``function_ref``'s enclave connection.

        Note that this context manager accomplishes the same functionality as
        :meth:`~Cape.connect`, except that it will also automatically
        :meth:`~Cape.close` the connection when exiting the context.

        **Usage** ::

            cape = Cape(url="https://app.capeprivacy.com")
            f = FunctionRef.from_json("function.json")

            with cape.function_context(f):

                c1 = cape.invoke(3, 4, use_serdio=True)
                print(c1)  # 5

                c2 = cape.invoke(5, 12, use_serdio=True)
                print(c2)  # 13

            # websocket connection is automatically closed

        Args:
            function_ref: A function ID or :class:`~.function_ref.FunctionRef`
                representing a deployed Cape function.

        Raises:
            RuntimeError: if the websocket response or the enclave attestation doc is
                malformed, or if the enclave fails to return a function checksum
                matching our own.
            Exception: if the enclave threw an error while trying to fulfill the
                connection request.
        """
        try:
            yield await self.connect(function_ref, pcrs)
        finally:
            await self.close()

    @_synchronizer
    async def invoke(
        self, *args: Any, serde_hooks=None, use_serdio: bool = False, **kwargs: Any
    ) -> Any:
        """Invokes a function call from the currently connected websocket.

        This method assumes that the client is currently maintaining an open websocket
        connection to an enclave hosting a particular Cape function. Care should be
        taken to ensure that the function_red that spawned the connection is the
        correct one. The connection should be closed with :meth:`~Cape.close` once the
        caller is finished with its invocations.

        Args:
            *args: Arguments to pass to the connected Cape function. If
                ``use_serdio=False``, we expect a single argument of type ``bytes``.
                Otherwise, these arguments should match the positional arguments
                of the undecorated Cape handler, and they will be auto-serialized by
                Serdio before being sent in the request.
            serde_hooks: An optional pair of serdio encoder/decoder hooks convertible
                to :class:`serdio.SerdeHookBundle`. The hooks are necessary if the
                ``args`` / ``kwargs`` have any user-defined types that can't be handled
                by vanilla Serdio. See :func:`serdio.bundle_serde_hooks` for supported
                types.
            use_serdio: Boolean controlling whether or not the inputs should be
                auto-serialized by serdio.
            kwargs: Keyword arguments to be passed to the connected Cape function.
                These are treated the same way as the ``args`` are.

        Returns:
            If ``use_serdio=True``, returns the auto-deserialized result of calling the
            connected Cape function on the given ``args`` / ``kwargs``.
            If ``use_serdio=False``, returns the output of the Cape function as raw
            bytes.

        Raises:
            RuntimeError: if serialized inputs could not be HPKE-encrypted, or if
                websocket response is malformed.
        """
        if serde_hooks is not None:
            serde_hooks = serdio.bundle_serde_hooks(serde_hooks)
        return await self._request_invocation(serde_hooks, use_serdio, *args, **kwargs)

    @_synchronizer
    async def key(
        self,
        *,
        username: Optional[str] = None,
        token: Optional[str] = None,
        key_path: Optional[Union[str, os.PathLike]] = None,
        pcrs: Optional[Dict[str, List[str]]] = None,
    ) -> bytes:
        """Load a Cape key from disk or download and persist an enclave-generated one.

        The caller must provide one of the following arguments: ``username``, ``token``,
        or ``key_path``.

        Args:
            username: An optional string representing the Github username of a Cape
                user. The resulting public key will be associated with their account,
                and data encrypted with this key will be available inside functions
                that user has deployed.
            token: An optional string representing a Cape authentication token. Usually
                retrieved from a :class:`~.function_ref.FunctionRef` via
                :attr:`FunctionRef.token`
            key_path: The path to the Cape key file. If the file already exists, the key
                will be read from disk and returned. Otherwise, a Cape key will be
                requested from the Cape platform and written to this location.
                If None, the default path is ``"$HOME/.config/cape/capekey.pub.der"``,
                or alternatively whatever path is specified by expanding the env
                variables ``CAPE_LOCAL_CONFIG_DIR / CAPE_LOCAL_CAPE_KEY_FILENAME``.
            pcrs: A dictionary of PCR indexes to a list of potential values.

        Returns:
            Bytes containing the Cape key. The key is also cached on disk for later
            use.

        Raises:
            RuntimeError: if the enclave attestation doc does not contain a Cape key,
                if the websocket response or the attestation doc is malformed.
            Exception: if the enclave threw an error while trying to fulfill the
                connection request.
        """
        if username is None and token is None and key_path is None:
            raise ValueError(
                "Must supply one of [`username`, `token`, `key_path`] arguments, but "
                "found `None` for all of them."
            )
        if username is not None and token is not None:
            raise ValueError(
                "Provided both `username` and `token` arguments, but these are "
                "mutually exclusive."
            )

        if key_path is None:
            config_dir = pathlib.Path(cape_config.LOCAL_CONFIG_DIR)
            key_qualifier = (
                username or token[-200:]
            )  # Shorten file name to avoid Errno 63 when saving file
            key_path = (
                config_dir
                / "encryption_keys"
                / key_qualifier
                / cape_config.LOCAL_CAPE_KEY_FILENAME
            )
        else:
            key_path = pathlib.Path(key_path)

        if key_path.exists():
            with open(key_path, "rb") as f:
                cape_key = f.read()
        else:
            if username is not None:
                cape_key = await self._request_key_with_username(username, pcrs=pcrs)
            else:
                cape_key = await self._request_key_with_token(token, pcrs=pcrs)
            await _persist_cape_key(cape_key, key_path)

        return cape_key

    @_synchronizer
    async def run(
        self,
        function_ref: fref.FunctionRef,
        *args: Any,
        pcrs: Optional[Dict[str, List[str]]] = None,
        serde_hooks=None,
        use_serdio: bool = False,
        **kwargs: Any,
    ) -> Any:
        """Single-shot version of connect + invoke + close.

        This method takes care of establishing a websocket connection via
        :meth:`~Cape.connect`, invoking it via :meth:`~Cape.invoke`, and then finally
        closing the connection with :meth:`~Cape.close`. This method should be
        preferred when the caller doesn't need to invoke a Cape function more than once.

        Args:
            function_ref: A :class:`~.function_ref.FunctionRef` representing
                a deployed Cape function.
            *args: Arguments to pass to the connected Cape function. If
                ``use_serdio=False``, we expect a single argument of type ``bytes``.
                Otherwise, these arguments should match the positional arguments
                of the undecorated Cape handler, and they will be auto-serialized by
                Serdio before being sent in the request.
            serde_hooks: An optional pair of serdio encoder/decoder hooks convertible
                to :class:`serdio.SerdeHookBundle`. The hooks are necessary if the
                ``args`` / ``kwargs`` have any user-defined types that can't be handled
                by vanilla Serdio. See :func:`serdio.bundle_serde_hooks` for supported
                types.
            use_serdio: Boolean controlling whether or not the inputs should be
                auto-serialized by serdio.
            kwargs: Keyword arguments to be passed to the connected Cape function.
                These are treated the same way as the ``args`` are.

        Returns:
            If ``use_serdio=True``, returns the auto-deserialized result of calling the
            connected Cape function on the given ``args`` / ``kwargs``.
            If ``use_serdio=False``, returns the output of the Cape function as raw
            bytes.

        Raises:
            RuntimeError: if serialized inputs could not be HPKE-encrypted, or if
                websocket response is malformed.
        """
        function_ref = _check_function_ref(function_ref)
        if serde_hooks is not None:
            serde_hooks = serdio.bundle_serde_hooks(serde_hooks)
        async with self.function_context(function_ref, pcrs):
            result = await self.invoke(
                *args, serde_hooks=serde_hooks, use_serdio=use_serdio, **kwargs
            )
        return result

    async def _request_connection(self, function_ref, pcrs=None):
        fn_endpoint = f"{self._url}/v1/run/{function_ref.id}"

        self._root_cert = self._root_cert or attest.download_root_cert()
        self._ctx = _EnclaveContext(
            endpoint=fn_endpoint,
            auth_protocol="cape.function",
            auth_token=function_ref.token,
            root_cert=self._root_cert,
        )
        attestation_doc = await self._ctx.bootstrap(pcrs)

        user_data = attestation_doc.get("user_data")
        checksum = function_ref.checksum
        if checksum is not None and user_data is None:
            # Close the connection explicitly before throwing exception
            await self._ctx.close()
            raise RuntimeError(
                f"No function checksum received from enclave, expected{checksum}."
            )

        user_data_dict = json.loads(user_data)
        received_checksum = user_data_dict.get("func_checksum")
        if checksum is not None:
            # Checksum is hex encoded, we manipulate it to string for comparison
            received_checksum = str(base64.b64decode(received_checksum).hex())
            if str(checksum) != str(received_checksum):
                # Close the connection explicitly before throwing exception
                await self._ctx.close()
                raise RuntimeError(
                    "Returned checksum did not match provided, "
                    f"got: {received_checksum}, want: {checksum}."
                )
        return

    async def _request_invocation(self, serde_hooks, use_serdio, *args, **kwargs):
        # If multiple args and/or kwargs are supplied to the Cape function through
        # Cape.run or Cape.invoke, before serialization, we pack them
        # into a dictionary with the following keys:
        # {"cape_fn_args": <tuple_args>, "cape_fn_kwargs": <dict_kwargs>}.
        single_input = _maybe_get_single_input(args, kwargs)
        if single_input is not None:
            inputs = single_input
        elif single_input is None and not use_serdio:
            raise ValueError(
                "Expected a single input of type 'bytes' when use_serdio=False.\n"
                "Found:"
                f"\t- args: {args}"
                f"\t- kwargs: {kwargs}"
            )

        if serde_hooks is not None:
            encoder_hook, decoder_hook = serde_hooks.unbundle()
            use_serdio = True
        else:
            encoder_hook, decoder_hook = None, None

        if use_serdio:
            inputs = serdio.serialize(*args, encoder=encoder_hook, **kwargs)

        if not isinstance(inputs, bytes):
            raise TypeError(
                f"The input type is: {type(inputs)}. Provide input as bytes or "
                "set use_serdio=True for PyCape to serialize your input "
                "with Serdio."
            )

        result = await self._ctx.invoke(inputs)

        if use_serdio:
            result = serdio.deserialize(result, decoder=decoder_hook)

        return result

    async def _request_key_with_username(
        self,
        username: str,
        pcrs: Optional[Dict[str, List[str]]] = None,
    ) -> bytes:
        user_key_endpoint = f"{self._url}/v1/user/{username}/key"
        response = requests.get(user_key_endpoint)
        adoc_blob = response.json()["attestation_document"]
        root_cert = self._root_cert or attest.download_root_cert()
        attestation_doc = attest.parse_attestation(
            base64.b64decode(adoc_blob), root_cert
        )
        if pcrs is not None:
            attest.verify_pcrs(pcrs, attestation_doc)

        user_data = attestation_doc.get("user_data")
        user_data_dict = json.loads(user_data)
        cape_key = user_data_dict.get("key")
        if cape_key is None:
            raise RuntimeError(
                "Enclave response did not include a Cape key in attestation user data."
            )
        return base64.b64decode(cape_key)

    async def _request_key_with_token(
        self,
        token: str,
        pcrs: Optional[Dict[str, List[str]]] = None,
    ) -> bytes:
        key_endpoint = f"{self._url}/v1/key"
        self._root_cert = self._root_cert or attest.download_root_cert()
        key_ctx = _EnclaveContext(
            key_endpoint,
            auth_protocol="cape.function",
            auth_token=token,
            root_cert=self._root_cert,
        )
        attestation_doc = await key_ctx.bootstrap(pcrs)
        await key_ctx.close()  # we have the attestation doc, no longer any need for ctx
        user_data = attestation_doc.get("user_data")
        user_data_dict = json.loads(user_data)
        cape_key = user_data_dict.get("key")
        if cape_key is None:
            raise RuntimeError(
                "Enclave response did not include a Cape key in attestation user data."
            )
        return base64.b64decode(cape_key)


class _EnclaveContext:
    """A context managing a connection to a particular enclave instance."""

    def __init__(self, endpoint, auth_protocol, auth_token, root_cert):
        self._endpoint = _transform_url(endpoint)
        self._auth_token = auth_token
        self._auth_protocol = auth_protocol
        self._root_cert = root_cert
        ssl_ctx = ssl.create_default_context()
        if cape_config.DEV_DISABLE_SSL:
            ssl_ctx.check_hostname = False
            ssl_ctx.verify_mode = ssl.CERT_NONE
        self._ssl_ctx = ssl_ctx

        # state to be explicitly created/destroyed by callers via bootstrap/close
        self._websocket = None
        self._public_key = None

    async def authenticate(self):
        nonce = _generate_nonce()
        request = _create_connection_request(nonce)
        _logger.debug("\n> Sending authentication request...")
        await self._websocket.send(request)
        _logger.debug("* Waiting for attestation document...")
        msg = await self._websocket.recv()
        _logger.debug("< Auth completed. Received attestation document.")
        return _parse_wss_response(msg)

    async def bootstrap(self, pcrs: Optional[Dict[str, List[str]]] = None):
        _logger.debug(f"* Dialing {self._endpoint}")
        self._websocket = await websockets.connect(
            self._endpoint,
            ssl=self._ssl_ctx,
            subprotocols=[self._auth_protocol, self._auth_token],
            max_size=None,
        )
        _logger.debug("* Websocket connection established")

        auth_response = await self.authenticate()
        attestation_doc = attest.parse_attestation(auth_response, self._root_cert)
        self._public_key = attestation_doc["public_key"]

        if pcrs is not None:
            attest.verify_pcrs(pcrs, attestation_doc)

        return attestation_doc

    async def close(self):
        if self._websocket is not None:
            await self._websocket.close()
            self._websocket = None
        self._public_key = None

    async def invoke(self, inputs: bytes) -> bytes:
        input_ciphertext = enclave_encrypt.encrypt(self._public_key, inputs)

        _logger.debug("> Sending encrypted inputs")
        try:
            await self._websocket.send(input_ciphertext)
        except websockets.exceptions.ConnectionClosedOK:
            await self.close()
            raise RuntimeError(
                "Enclave websocket connection was closed, likely due to timeout error. "
                "Please invoke your function more frequently to keep the connection "
                "alive for more than 60 seconds."
            )

        invoke_response = await self._websocket.recv()
        _logger.debug("< Received function results")

        return _parse_wss_response(invoke_response)


# TODO What should be the length?
def _generate_nonce(length=8):
    """
    Generates a string of digits between 0 and 9 of a given length
    """
    nonce = "".join([str(random.randint(0, 9)) for i in range(length)])
    _logger.debug(f"* Generated nonce: {nonce}")
    return nonce


def _create_connection_request(nonce):
    """
    Returns a json string with nonce
    """
    request = {"message": {"nonce": nonce}}
    return json.dumps(request)


def _parse_wss_response(response):
    """
    Returns the inner message field received in a WebSocket message from enclave
    """
    response = json.loads(response)
    if "error" in response:
        raise Exception(response["error"])
    response_msg = _handle_expected_field(
        response,
        "message",
        fallback_err="Missing 'message' field in websocket response.",
    )
    inner_msg = _handle_expected_field(
        response_msg,
        "message",
        fallback_err=(
            "Malformed websocket response contents: missing inner 'message' field."
        ),
    )
    return base64.b64decode(inner_msg)


def _handle_expected_field(dictionary, field, *, fallback_err=None):
    """
    Returns value of a provided key from dictionary, optionally raising
    a custom RuntimeError if it's missing.
    """
    v = dictionary.get(field, None)
    if v is None:
        if fallback_err is not None:
            _logger.error(fallback_err)
            raise RuntimeError(fallback_err)
        raise RuntimeError(f"Dictionary {dictionary} missing key {field}.")
    return v


def _check_function_ref(function_ref):
    """
    Check if the FunctionRef object contains a function ID and token.
    """
    if isinstance(function_ref, fref.FunctionRef):
        if function_ref.id is None:
            raise ValueError("The function ID in the provided FunctionRef is empty.")

        if function_ref.token is None:
            raise ValueError("The function token in the provided FunctionRef is empty.")

        return function_ref
    else:
        raise TypeError(
            "`function_ref` arg must be a FunctionRef object, "
            f" found {type(function_ref)}."
        )


def _maybe_get_single_input(args, kwargs):
    single_arg = len(args) == 1 and len(kwargs) == 0
    single_kwarg = len(args) == 0 and len(kwargs) == 1
    if single_arg:
        return args[0]
    elif single_kwarg:
        return kwargs.items()[0][1]


async def _persist_cape_key(cape_key: str, key_path: pathlib.Path):
    key_path.parent.mkdir(parents=True, exist_ok=True)
    with open(key_path, "wb") as f:
        f.write(cape_key)


def _transform_url(url):
    url = urllib.parse.urlparse(url)
    if url.scheme == "https":
        return url.geturl().replace("https://", "wss://")
    elif url.scheme == "http":
        return url.geturl().replace("http://", "ws://")
    return url.geturl()



import base64

from pycape._enclave_encrypt import encrypt


class TestEnclaveEncryption:
    def test_data_encryption(self):
        plaintext = b"private_data"

        public_key = base64.b64decode("d4Y4fxNr/hga+d86m2Lw+SXu+QO6Uuk3yrtrS9CoVgI=")

        _ = encrypt(public_key, plaintext)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../../"))


# -- Project information -----------------------------------------------------

project = "PyCape"
copyright = "2022, Cape Privacy"
author = "Cape Privacy"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "myst_parser",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

napoleon_include_init_with_doc = True
myst_heading_anchors = 3

# Add any paths that contain templates here, relative to this directory.
templates_path = ["../_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["../_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "logo_only": True,
    "show_toc_level": 2,
    "repository_url": "https://github.com/capeprivacy/pycape",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "../_static/cape_logo.png"
html_favicon = "../_static/favicon.png"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["../_static"]



import asyncio
import os
import pathlib

import pycape
from pycape.experimental import cli


async def main(cape: pycape.Cape, deploy_path: os.PathLike, echo: str) -> str:
    function_ref = await cli.deploy(deploy_path)
    echo_arg = echo.encode()
    echo_enc = await cape.encrypt(echo_arg, token=function_ref.token)
    async with cape.function_context(function_ref):
        result = await cape.invoke(echo_enc)
    return result.decode()


if __name__ == "__main__":
    echo_path = pathlib.Path(__file__).parent.absolute() / "echo"

    cape = pycape.Cape()
    echo = asyncio.run(main(cape, echo_path, "Welcome to Cape."))
    print(echo)



import os
import pathlib

import pycape._config as modself

# the following config vars will be set for this module, unless overridden
# by an identical env variable prefixed w/ "CAPE_*", e.g. CAPE_DEV_DISABLE_SSL=True
_CAPE_ENVVAR_DEFAULTS = {
    "DEV_DISABLE_SSL": False,
    "ENCLAVE_HOST": "https://app.capeprivacy.com",
    "LOCAL_AUTH_FILENAME": "auth",
    "LOCAL_CAPE_KEY_FILENAME": "capekey.pub.der",
    "LOCAL_CONFIG_DIR": str(pathlib.Path.home() / ".config" / "cape"),
}


def _init_config():
    for cfgvar, default in _CAPE_ENVVAR_DEFAULTS.items():
        envname = "_".join(["CAPE", cfgvar])
        envvar = os.environ.get(envname)
        if envvar is None:
            envvar = default
        setattr(modself, cfgvar, envvar)


_init_config()



import os
import pathlib

from pycape import Cape
from pycape import FunctionRef

if __name__ == "__main__":
    url = os.environ.get("CAPE_HOST", "https://app.capeprivacy.com")
    function_json = os.environ.get("FUNCTION_JSON", "echo_token.json")
    function_json = pathlib.Path(__file__).parent.absolute() / function_json

    function_ref = FunctionRef.from_json(function_json)
    cape = Cape(url=url)
    cape.connect(function_ref)
    result = cape.invoke("Hello Cape".encode())
    print(f"The result is: {result.decode()}")
    result = cape.invoke("Hello Gavin".encode())
    print(f"The result is: {result.decode()}")
    result = cape.invoke("Hello Hello".encode())
    print(f"The result is: {result.decode()}")
    cape.close()

    # Note that instead you can use the connection_context
    # context manager which will automatically close the
    # connection and reset websocket connection's states.
    with cape.function_context(function_ref):
        result = cape.invoke("Hello Cape".encode())
        print(f"The result is: {result.decode()}")
        result = cape.invoke("Hello Gavin".encode())
        print(f"The result is: {result.decode()}")
        result = cape.invoke("Hello Hello".encode())
        print(f"The result is: {result.decode()}")



import os
import pathlib

from pycape import Cape
from pycape import FunctionRef

if __name__ == "__main__":
    url = os.environ.get("CAPE_HOST", "https://app.capeprivacy.com")
    function_json = os.environ.get("FUNCTION_JSON", "echo_token.json")
    function_json = pathlib.Path(__file__).parent.absolute() / function_json

    function_ref = FunctionRef.from_json(function_json)
    cape = Cape(url=url)
    input = "Welcome to Cape".encode()
    result = cape.run(function_ref, input)

    print(f"The result is: {result.decode()}")



"""Utility functions supporting the Cape encrypt functionality."""
import os
from typing import Tuple

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.ciphers import aead


def encrypt(message: bytes, key: bytes) -> bytes:
    """Encrypt a ``message`` with a Cape ``key``.

    This function uses envelope encryption. The message is first AES-encrypted with an
    ephemeral AES key, and then this key is itself encrypted with a given RSA public
    key.

    Args:
        message: Bytes to encrypt.
        key: Bytes representing the Cape key. Needs to be a valid, DEM-encoded RSA
            public key.

    Returns:
        Bytes represeting the encryption of ``message``. The bytes are a concatenation
        of the AES-ciphertext of ``message``, an AES nonce, and the RSA-ciphertext of
        the AES key.

    Raises:
        ValueError: if the ``key`` is not a valid DEM-encoded RSA public key.
    """
    # cape key is DEM-encoded RSA key
    rsa_key = _parse_rsa_key(key)
    # create ephemeral AES key
    aes_key = _aes_keygen(256)
    # encrypt message w/ AES
    data_ctxt, nonce = _aes_encrypt(message, aes_key)
    # encrypt the ephemeral AES key w/ RSA
    key_ctxt = _rsa_encrypt(aes_key, rsa_key)
    # concatenate everything into a single bytes obj
    final_ctxt = key_ctxt + nonce + data_ctxt
    return final_ctxt


def _aes_encrypt(inputs: bytes, key: bytes) -> Tuple[bytes, bytes]:
    encryptor = aead.AESGCM(key)
    nonce = os.urandom(12)  # AESGCM nonce size is 12
    ctxt = encryptor.encrypt(nonce, inputs, None)
    return ctxt, nonce


def _aes_keygen(bitlength: int) -> bytes:
    return aead.AESGCM.generate_key(bitlength)


def _parse_rsa_key(key: bytes) -> rsa.RSAPublicKey:
    public_key = serialization.load_der_public_key(key)
    if not isinstance(public_key, rsa.RSAPublicKey):
        raise ValueError(
            f"Decoded 'key' expected to be RSAPublicKey, found {type(public_key)}"
        )
    return public_key


def _rsa_encrypt(inputs: bytes, public_key: rsa.RSAPublicKey) -> bytes:
    return public_key.encrypt(
        inputs,
        padding=padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )



import base64
import json
import unittest

from pycape.cape import _create_connection_request
from pycape.cape import _generate_nonce
from pycape.cape import _handle_expected_field
from pycape.cape import _parse_wss_response


class TestCape(unittest.TestCase):
    def test_generate_nonce(self):
        length = 8
        nonce = _generate_nonce(length=length)
        self.assertTrue(isinstance(nonce, str))
        self.assertTrue(len(nonce), length)

    def test_create_connection_request(self):
        nonce = "90444145"
        conn_req = _create_connection_request(nonce)
        self.assertEqual(conn_req, json.dumps({"message": {"nonce": "90444145"}}))

    def test_handle_expected_field(self):
        response = '{"message": "connected"}'
        response = json.loads(response)
        response_msg = _handle_expected_field(
            response,
            "message",
            fallback_err="Missing 'message' field in websocket response.",
        )
        self.assertEqual(response_msg, "connected")

    def test_parse_wss_response(self):
        response = json.dumps({"message": {"message": "conn"}})
        inner_msg = _parse_wss_response(response)
        self.assertEqual(inner_msg, base64.b64decode("conn"))


if __name__ == "__main__":
    unittest.main()



import statistics

import serdio


@serdio.lift_io(as_handler=True)
def cape_handler(x):
    return statistics.mean(x)


# NOTE: this would have also worked, since `compute_mean` is trivial
#  cape_handler = serdio.lift_io(statistics.mean).as_cape_handler()



import base64
import datetime
import json
import time

import cbor2
import pytest
from cose.algorithms import Es384
from cose.keys import EC2Key
from cose.keys.curves import P384
from cose.messages import Sign1Message
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509.oid import NameOID
from OpenSSL import crypto

from pycape import _attestation as attest

root_subject = issuer = x509.Name(
    [
        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Texas"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "Austin"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "My Company"),
        x509.NameAttribute(NameOID.COMMON_NAME, "My CA"),
    ]
)

intermediate_subject = issuer = x509.Name(
    [
        x509.NameAttribute(NameOID.COUNTRY_NAME, "JP"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Tokyo"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "Tokyo"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Cool Comp"),
        x509.NameAttribute(NameOID.COMMON_NAME, "COOL"),
    ]
)

cert_subject = x509.Name(
    [
        x509.NameAttribute(NameOID.COUNTRY_NAME, "CA"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Nova Scotia"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "Halifax"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "New Org Name!"),
    ]
)


class TestAttestation:
    def test_parse_attestation(self):
        crv = P384
        root_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )
        private_key = ec.generate_private_key(crv.curve_obj, backend=default_backend())

        root_cert = create_root_cert(root_private_key, root_subject)
        intermediate_cert = create_child_cert(
            root_cert, root_private_key, root_private_key, intermediate_subject, ca=True
        )
        cert = create_child_cert(
            intermediate_cert, root_private_key, private_key, cert_subject, ca=False
        )
        doc_bytes = create_attestation_doc(intermediate_cert, cert)
        attestation = create_cose_1_sign_msg(doc_bytes, private_key)

        attestation_doc = attest.parse_attestation(
            attestation, root_cert.public_bytes(Encoding.PEM)
        )
        public_key = attestation_doc["public_key"]
        user_data = attestation_doc.get("user_data")
        expected_user_data = json.dumps({"func_hash": "stuff"})
        assert user_data == expected_user_data
        assert len(public_key) == 32

    def test_verify_attestation_signature(self):
        crv = P384
        root_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )
        private_key = ec.generate_private_key(crv.curve_obj, backend=default_backend())

        root_cert = create_root_cert(root_private_key, root_subject)
        intermediate_cert = create_child_cert(
            root_cert, root_private_key, root_private_key, intermediate_subject, ca=True
        )
        cert = create_child_cert(
            intermediate_cert, root_private_key, private_key, cert_subject, ca=False
        )
        doc_bytes = create_attestation_doc(intermediate_cert, cert)
        attestation = create_cose_1_sign_msg(doc_bytes, private_key)

        payload = cbor2.loads(attestation)
        doc = cbor2.loads(payload[2])

        attest.verify_attestation_signature(attestation, doc["certificate"])

    def test_verify_cert_chain(self):
        crv = P384
        root_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )
        private_key = ec.generate_private_key(crv.curve_obj, backend=default_backend())

        root_cert = create_root_cert(root_private_key, root_subject)
        intermediate_cert = create_child_cert(
            root_cert, root_private_key, root_private_key, intermediate_subject, ca=True
        )
        cert = create_child_cert(
            intermediate_cert, root_private_key, private_key, cert_subject
        )

        doc_bytes = create_attestation_doc(intermediate_cert, cert)
        attestation = create_cose_1_sign_msg(doc_bytes, private_key)

        payload = cbor2.loads(attestation)
        doc = cbor2.loads(payload[2])

        root_cert_pem = root_cert.public_bytes(Encoding.PEM)

        attest.verify_cert_chain(root_cert_pem, doc["cabundle"], doc["certificate"])

    def test_verify_cert_chain_fails_if_bad_intermediate(self):
        crv = P384
        root_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )
        intermediate_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )

        private_key = ec.generate_private_key(crv.curve_obj, backend=default_backend())

        root_cert = create_root_cert(root_private_key, root_subject)
        intermediate_cert = create_root_cert(
            intermediate_private_key, intermediate_subject
        )
        cert = create_child_cert(
            intermediate_cert, root_private_key, private_key, cert_subject, ca=False
        )

        doc_bytes = create_attestation_doc(intermediate_cert, cert)
        attestation = create_cose_1_sign_msg(doc_bytes, private_key)

        payload = cbor2.loads(attestation)
        doc = cbor2.loads(payload[2])

        root_cert_pem = root_cert.public_bytes(Encoding.PEM)

        with pytest.raises(crypto.X509StoreContextError):
            attest.verify_cert_chain(root_cert_pem, doc["cabundle"], doc["certificate"])

    def test_verify_pcrs(self):
        crv = P384
        root_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )
        private_key = ec.generate_private_key(crv.curve_obj, backend=default_backend())

        root_cert = create_root_cert(root_private_key, root_subject)
        intermediate_cert = create_child_cert(
            root_cert, root_private_key, root_private_key, intermediate_subject, ca=True
        )
        cert = create_child_cert(
            intermediate_cert, root_private_key, private_key, cert_subject, ca=False
        )
        doc_bytes = create_attestation_doc(intermediate_cert, cert)
        attestation = create_cose_1_sign_msg(doc_bytes, private_key)

        attestation_doc = attest.parse_attestation(
            attestation, root_cert.public_bytes(Encoding.PEM)
        )

        attest.verify_pcrs({"0": [b"pcrpcrpcr".hex()]}, attestation_doc)

    def test_verify_pcrs_fail(self):
        crv = P384
        root_private_key = ec.generate_private_key(
            crv.curve_obj, backend=default_backend()
        )
        private_key = ec.generate_private_key(crv.curve_obj, backend=default_backend())

        root_cert = create_root_cert(root_private_key, root_subject)
        intermediate_cert = create_child_cert(
            root_cert, root_private_key, root_private_key, intermediate_subject, ca=True
        )
        cert = create_child_cert(
            intermediate_cert, root_private_key, private_key, cert_subject, ca=False
        )
        doc_bytes = create_attestation_doc(intermediate_cert, cert)
        attestation = create_cose_1_sign_msg(doc_bytes, private_key)

        attestation_doc = attest.parse_attestation(
            attestation, root_cert.public_bytes(Encoding.PEM)
        )

        with pytest.raises(Exception):
            attest.verify_pcrs({"0": [b"pcrpcr".hex()]}, attestation_doc)


def create_cose_1_sign_msg(payload, private_key):
    crv = P384
    d_value = private_key.private_numbers().private_value
    x_coor = private_key.public_key().public_numbers().x
    y_coor = private_key.public_key().public_numbers().y

    key = EC2Key(
        crv=crv,
        d=d_value.to_bytes(crv.size, "big"),
        x=x_coor.to_bytes(crv.size, "big"),
        y=y_coor.to_bytes(crv.size, "big"),
    )

    phdr = {1: Es384}
    msg = Sign1Message(phdr=phdr, payload=payload, key=key)

    sig = msg.compute_signature()
    msg._signature = sig

    return msg.encode(tag=False)


def create_attestation_doc(intermediate_cert, cert):
    cert = cert.public_bytes(Encoding.DER)
    intermediate_cert = intermediate_cert.public_bytes(Encoding.DER)
    user_data = json.dumps({"func_hash": "stuff"})
    public_key = base64.b64decode("d4Y4fxNr/hga+d86m2Lw+SXu+QO6Uuk3yrtrS9CoVgI=")
    obj = {
        "module_id": "my-module",
        "timestamp": time.gmtime(),
        "digest": "abcd",
        "pcrs": {0: b"pcrpcrpcr"},
        "certificate": cert,
        "cabundle": [intermediate_cert],
        "public_key": public_key,
        "user_data": user_data,
    }

    return cbor2.encoder.dumps(obj)


def create_root_cert(root_key, subject):
    issuer = subject
    root_cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(root_key.public_key())
        .serial_number(x509.random_serial_number())
        .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=True)
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=3650))
        .sign(root_key, hashes.SHA256(), default_backend())
    )

    return root_cert


def create_child_cert(parent_cert, parent_key, cert_key, subject, ca=False):
    builder = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(parent_cert.issuer)
        .public_key(cert_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=30))
    )

    if ca:
        builder = builder.add_extension(
            x509.BasicConstraints(ca=True, path_length=None), critical=True
        )

    cert = builder.sign(parent_key, hashes.SHA256(), default_backend())

    return cert
